package Time::Moment::Role::BOKUTIN;
$Time::Moment::Role::BOKUTIN::VERSION = '0.17';
# コンストラクタ系はオーバーライドしない。オフセットもそのまま。
# 読み出しのメソッドはオーバーライドも可。strftime の %a はローカル。

use utf8;
use strict;
use warnings;
use feature qw(:5.10);
use Role::Tiny;
with "Time::Moment::Role::TimeZone::Fast";

use Carp qw(croak cluck);
use POSIX qw(:locale_h);
use Scalar::Util qw(blessed);
use Time::TZOffset qw(tzoffset_as_seconds);

use namespace::clean;

# 曖昧さが無さそうな文字列をよしなに解釈します。
# offset が指定されない場合は
#   epoch の場合はUTCとして
#   文字列,rd の場合はローカル時間として
# 解釈し、localtime のオフセットを付加します。
# 毎回 offset を計算するコスト (man 3 localtime を呼ぶコストは) は支払います。
sub from_fuzzy {
    my ($class, $value) = (shift, shift);

    local $@;
    my $tm = eval {
        if ( blessed $value ) {
            my $obj = $value;
            if ($obj->isa('DateTime')) {
                if ($obj->time_zone->is_floating) {
                    $obj = $obj->clone;
                    $obj->set_time_zone('local');
                }
            }
            return $class->from_object($obj);
        }
        elsif ( $value eq 'from_object' ) {
            my $obj = $_[0];
            if ($obj->isa('DateTime')) {
                if ($obj->time_zone->is_floating) {
                    $obj = $obj->clone;
                    $obj->set_time_zone('local');
                }
            }
            return $class->from_object($obj);
        }
        elsif ( $value eq 'from_string' ) {
            # from_string は offset の無い文字列は許可していないので、必ず offset を指定していることになるはず
            return $class->from_string(@_);
        }
        elsif ( $value eq 'from_epoch' ) {
            return $class->from_epoch(@_)->with_system_offset_same_instant;
        }
        elsif ( $value eq 'from_rd' ) {
            my ($num, %p) = @_;
            my $tm = $class->from_rd(@_);
            unless (length $p{offset}) {
                $tm = $tm->with_system_offset_same_local;
            }
            return $tm;
        }
        elsif ( $value =~ /^from_/ ) { # from_jd|from_mjd
            $class->$value(@_)->with_system_offset_same_instant;
        }
        elsif ( length $value == 10 and ( $value =~ /^(\d{4})-(\d{2})-(\d{2})$/ or $value =~ m{^(\d{4})/(\d{2})/(\d{2})$} ) ) {
            my ($year, $month, $day) = @{^CAPTURE};
            my @lt = ( 0, 0, 0, $day, $month-1, $year-1900 );
            my $offset = tzoffset_as_seconds(@lt)/60;
            return $class->new( year => $1, month => $2, day => $3, offset => $offset );
        }
        elsif ( length $value == 19 and $value =~ /^(\d{4})-(\d{2})-(\d{2})[T ](\d{2}):(\d{2}):(\d{2})$/ ) {
            my ($year, $month, $day, $hour, $minute, $second) = @{^CAPTURE};
            my @lt = ( $second, $minute, $hour, $day, $month-1, $year-1900 );
            my $offset = tzoffset_as_seconds(@lt)/60;
            return $class->new( year => $year, month => $month, day => $day, hour => $hour, minute => $minute, second => $second, offset => $offset );
        }
        # wordpress wp.getPost post_modified 20161125T20:40:17
        elsif ( length $value == 17 and $value =~ /^(\d{4})(\d{2})(\d{2})[T ](\d{2}):(\d{2}):(\d{2})$/ ) {
            my ($year, $month, $day, $hour, $minute, $second) = @{^CAPTURE};
            my @lt = ( $second, $minute, $hour, $day, $month-1, $year-1900 );
            my $offset = tzoffset_as_seconds(@lt)/60;
            return $class->new( year => $year, month => $month, day => $day, hour => $hour, minute => $minute, second => $second, offset => $offset );
        }
        elsif ( length $value > 20 and $value =~ /^(\d{4})-(\d{2})-(\d{2})[T ](\d{2}):(\d{2}):(\d{2})\.(\d+)$/ ) {
            my ($year, $month, $day, $hour, $minute, $second) = @{^CAPTURE};
            my $nanosecond = sprintf "%09d", $7;
            my @lt = ( $second, $minute, $hour, $day, $month-1, $year-1900 );
            my $offset = tzoffset_as_seconds(@lt)/60;
            return $class->new( year => $year, month => $month, day => $day, hour => $hour, minute => $minute, second => $second, nanosecond => $nanosecond, offset => $offset );
        }
        # 2009-06-10T02:28:00+09:00
        elsif ( length $value == 25 and $value =~ /^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})([-+])(\d{2}):(\d{2})$/ ) {
            my ($year, $month, $day, $hour, $minute, $second) = @{^CAPTURE};
            my $offset = ($8*60 + $9) * "${7}1";
            return $class->new( year => $year, month => $month, day => $day, hour => $hour, minute => $minute, second => $second, offset => $offset );
        }
        # 2014-11-13T10:00:13.833253+09:00
        elsif ( length $value >= 27 and $value =~ /^(\d{4})-(\d{2})-(\d{2})[T ](\d{2}):(\d{2}):(\d{2})\.(\d+)([-+])(\d{2}):(\d{2})$/ ) {
            my ($year, $month, $day, $hour, $minute, $second) = @{^CAPTURE};
            my $nanosecond = sprintf "%09d", $7;
            my $offset = ($9*60 + $10) * "${8}1";
            return $class->new( year => $year, month => $month, day => $day, hour => $hour, minute => $minute, second => $second, nanosecond => $nanosecond, offset => $offset );
        }
        elsif ( length $value == 7 and $value =~ /^(\d{4})-(\d{2})$/ ) {
            my ($year, $month) = @{^CAPTURE};
            my @lt = ( 0, 0, 0, 1, $month-1, $year-1900 );
            my $offset = tzoffset_as_seconds(@lt)/60;
            return $class->new( year => $year, month => $month, offset => $offset );
        }
        elsif ( length $value == 8 and $value =~ /^(\d{4})(\d{2})(\d{2})$/ ) {
            my ($year, $month, $day) = @{^CAPTURE};
            my @lt = ( 0, 0, 0, $day, $month-1, $year-1900 );
            my $offset = tzoffset_as_seconds(@lt)/60;
            return $class->new( year => $year, month => $month, day => $day, offset => $offset );
        }
        # 2015年 8月 31日
        elsif ( length $value and $value =~ /^(\d+)年\s*(\d{1,2})月\s*(\d{1,2})日$/) {
            my ($year, $month, $day) = map $_+0, @{^CAPTURE};
            my @lt = ( 0, 0, 0, $day, $month, $year );
            my $offset = tzoffset_as_seconds(@lt)/60;
            return $class->new( year => $year, month => $month, day => $day, offset => $offset );
        }
        elsif ( length $value and $value eq 'infinity' or $value eq '-infinity' ) {
            die "Invalid value '$value'. INFINITE is not supported.";
        }
        elsif ( length $value and $value =~ /^\d+$/ ) {
            return $class->from_epoch($value)->with_system_offset_same_instant;
        }
        else {
            # from_string は offset の無い文字列は許可していないので、必ず offset を指定していることになるはず
            return $class->from_string($value, lenient => 1);
        }
    };

    if ($@) {
        my $i = 0;
        my $pad = join " ", map "arg@{[ $i++ ]}:$_", ($class,$value,@_);
        if ( !ref $@ and $@ =~ m{ at /} ) {
            croak $@ =~ s{(?= at /)}{ $pad}r;
        }
        else {
            cluck "invalid args. $pad";
            croak $@;
        }
    }

    $tm;
}

# deprecated
sub from_epoch_with_offset {
    my $class = shift;
    $class->from_epoch(@_)->with_system_offset_same_instant;
}

sub strptime {
    my ($class, $format, $string) = @_;

    require Time::Strptime;
    my ($epoch, $offset) = Time::Strptime::strptime($format, $string);
    $class->from_epoch($epoch)->with_offset_same_instant($offset/60);
}

{
    no strict 'refs';
    for my $h (0..24) {
        my $n = $h*60*60*1_000_000_000;
        *{"at_$h"} = sub { $_[0]->with_nanosecond_of_day($n) };
    }
}

sub yyyymm { $_[0]->year*100 + $_[0]->month }

sub ymd {
    my ($self, $sep) = @_;

    if (defined $sep) {
        $self->strftime("%Y$sep%m$sep%d");
    }
    else {
        $self->strftime('%F');
    }
}

sub ym {
    my ($self, $sep) = @_;

    if (defined $sep) {
        $self->strftime("%Y${sep}%m");
    }
    else {
        $self->year*100 + $self->month;
    }
}

sub hms {
    my ($self, $sep) = @_;

    if (defined $sep) {
        $self->strftime("%H${sep}%M${sep}%S");
    }
    else {
        $self->strftime('%X');
    }
}

sub day { shift->day_of_month(@_) }

sub local_rd_day { ($_[0]->local_rd_values)[0] }

sub dow { $_[0]->day_of_week }

sub day_name { $_[0]->strftime('%A') }

sub day_abbr { $_[0]->strftime('%a') }

sub render_as_mysql { $_[0]->strftime("%Y-%m-%d %H:%M:%S") }

sub for_pg { $_[0]."" }

# ひとまずは曜日のみ
around strftime => sub {
    my ($orig, $self, $format, $locale_code) = @_;

    my $ret = $self->$orig($format);

    return $ret unless $format =~ /\%[aA]/;

    $locale_code ||= setlocale(LC_TIME);

    return $ret if $locale_code =~ /^(?:c|en)\b/i;

    my $map = do {
        state $memo = {};
        $memo->{$locale_code} ||= do {
            require DateTime::Locale;
            my $loc = DateTime::Locale->load($locale_code);
            my %map;
            @map{qw(Monday Tuesday Wednesday Thursday Friday Saturday Sunday)} = @{$loc->day_format_wide};
            @map{qw(Mon    Tue     Wed       Thu      Fri    Sat      Sun   )} = @{$loc->day_format_narrow};
            \%map;
        };
    };
    $ret =~ s/(Monday|Mon|Tuesday|Tue|Wednesday|Wed|Thursday|Thu|Friday|Fri|Saturday|Sat|Sunday|Sun)/$map->{$1}/e;
    $ret;
};

sub tv_interval {
    my ($tm1, $tm2) = @_;

    my $cmp = $tm1->compare($tm2);
    if ( $cmp < 0 ) {
        Time::HiRes::tv_interval( [$tm1->epoch,$tm1->millisecond], [$tm2->epoch,$tm2->millisecond] );
    }
    elsif ( $cmp == 0 ) {
        0;
    }
    else {
        Time::HiRes::tv_interval( [$tm2->epoch,$tm2->millisecond], [$tm1->epoch,$tm1->millisecond] );
    }
}

sub is_holiday {
    my ($self) = @_;
    require Calendar::Japanese::Holiday;
    Calendar::Japanese::Holiday::isHoliday($self->year, $self->month, $self->day);
}

sub is_business_day {
    my ($self) = @_;

    return 0 if $self->day_of_week >= 6;
    return 0 if $self->is_holiday;
    return 1;
}

sub japanese_era {
    my ($tm) = @_;
    do { require Date::Japanese::Era; Date::Japanese::Era->import } unless $INC{"Date/Japanese/Era.pm"};
    Date::Japanese::Era->new($tm->year, $tm->month, $tm->day);
}

sub gengou_year {
    my ($tm) = @_;
    my $era = $tm->japanese_era;
    $era->gengou.$era->year;
}

sub delta_datetime_duration {
    my ($tm1, $tm2) = @_;

    my @units = qw(years months weeks days hours minutes seconds nanoseconds);

    my @args;
    my $cur = $tm1;
    for my $unit (@units) {
        my $d = $cur->${\"delta_${unit}"}($tm2);
        push @args, $unit => $d;
        $cur = $cur->${\"plus_${unit}"}($d);
    }
    require DateTime::Duration;
    DateTime::Duration->new(@args);
}

#
# BEGIN compat datetime
#
sub clone { $_[0] }
sub add {
    my $self = shift;
    die unless @_%2==0;

    while (@_) {
        my ($key, $value) = (shift, shift);
        $self = $self->${ \"plus_$key" }($value);
    }

    $self;
}
sub set {
    my $self = shift;
    die unless @_%2==0;

    my %map = (
        year  => "with_year",
        month => "with_month",
        day   => "with_day_of_month",
    );

    while (@_) {
        my ($key, $value) = (shift, shift);
        my $meth = $map{$key} || "with_$key";
        $self = $self->${ \$meth }($value);
    }

    $self;
}
sub set_locale {
    my ($self, $name) = @_;
    #$self->{_locale} = $name;
    $self;
}
sub set_time_zone {
    my ($self, $tz) = @_;

    unless ( blessed $tz ) {
        require DateTime::TimeZone;
        $tz = DateTime::TimeZone->new( name => $tz );
    }

    $self->with_time_zone_offset_same_instant($tz);
}
#
# END compat datetime
#

#
# BEGIN compat DBIx::Class::InflateColumn::DateTime and DateTime::Format::Pg
#
sub is_infinite { 0 }
#
# END compat datetime
#

around from_object => sub {
    my ($orig, $self) = (shift, shift);
    my $class = ref($self) || $self;
    bless $self->$orig(@_), $class;
};

sub delta_as_datetime_duration {
    my ($tm1, $tm2) = @_;

    my @units = qw(years months weeks days hours minutes seconds nanoseconds);

    my @args;
    my $cur = $tm1;
    for my $unit (@units) {
        my $d = $cur->${\"delta_${unit}"}($tm2);
        push @args, $unit => $d;
        $cur = $cur->${\"plus_${unit}"}($d);
    }
    require DateTime::Duration;
    DateTime::Duration->new(@args);
}

1;


1;

__END__

=pod

=encoding UTF-8

=head1 NAME

Time::Moment::Role::BOKUTIN

=cut
